#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Global Benchmark Database (GBD)
# Copyright (C) 2020 Markus Iser, Luca Springer, Karlsruhe Institute of Technology (KIT)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import logging
import os
import argparse

import werkzeug

from gbd_tool.util import eprint
from os.path import basename

import gbd_server

import tatsu
import flask
from flask import Flask, request, send_file, json, Response
from flask import render_template
from flask.logging import default_handler
from gbd_tool.gbd_api import GbdApi
from tatsu import exceptions
from werkzeug.middleware.proxy_fix import ProxyFix

app = Flask(__name__)

@app.route("/", methods=['GET'])
def quick_search():
    return render_template('index.html')

@app.route("/results", methods=['POST'])
def quick_search_results():
    with GbdApi(app.config['database']) as gbd_api:
        query = request.form.get('query')
        selected_features = list(filter(lambda x: x != '', request.form.get('selected_features').split(',')))
        if not len(selected_features):
            selected_features.append("filename")
        available_features = sorted(gbd_api.get_features())
        available_features.remove("local")
        features = sorted(list(set(available_features) & set(selected_features)))
        try:
            rows = list(gbd_api.query_search(query, features))
            features.insert(0, "GBDhash")
            result = list(dict((features[index], row[index]) for index in range(0, len(features))) for row in rows)
            return Response(json.dumps(result), status=200, mimetype="application/json")
        except tatsu.exceptions.FailedParse:
            return Response("Malformed query", status=400, mimetype="text/plain")
        except ValueError:
            return Response("Attribute not Available", status=400, mimetype="text/plain")


@app.route("/getdatabases", methods=["GET"])
def get_databases():
    with GbdApi(app.config['database']) as gbd_api:
        return json.dumps(gbd_api.get_databases())

@app.route('/getfeatures', defaults={'database': None})
@app.route('/getfeatures/<database>')
def get_features(database):
    with GbdApi(app.config['database']) as gbd_api:
        if database is None:
            available_features = sorted(gbd_api.get_features())
            available_features.remove("local")
            return Response(json.dumps(available_features), status=200, mimetype="application/json")
        elif database not in gbd_api.get_databases():
            return Response("Database does not exist in the running instance of GBD server", status=404,
                            mimetype="text/plain")
        else:
            return gbd_api.get_features(database)


@app.route("/exportcsv", methods=['POST'])
def get_csv_file():
    with GbdApi(app.config['database']) as gbd_api:
        query = request.form.get('query')
        selected_features = list(filter(lambda x: x != '', request.form.get('selected_features').split(',')))
        if not len(selected_features):
            selected_features.append("filename")
        results = gbd_api.query_search(query, selected_features)
        headers = ["hash"] + selected_features
        content = "\n".join([" ".join([str(entry) for entry in result]) for result in results])
        app.logger.info('Sending CSV file to {}'.format(request.remote_addr))
        file_name = "query_result.csv"
        return Response(" ".join(headers) + "\n" + content, mimetype='text/csv',
                        headers={"Content-Disposition": "attachment; filename=\"{}\"".format(file_name),
                                "filename": "{}".format(file_name)})


@app.route("/getinstances", methods=['POST'])
def get_url_file():
    with GbdApi(app.config['database']) as gbd_api:
        query = request.form.get('query')
        result = gbd_api.query_search(query, ["filename"])
        content = "\n".join([flask.url_for("get_file", hashvalue=row[0], filename=row[1], _external=True) for row in result])
        app.logger.info('Sending URL file to {}'.format(request.remote_addr))
        file_name = "query_result.uri"
        return Response(content, mimetype='text/uri-list',
                        headers={"Content-Disposition": "attachment; filename=\"{}\"".format(file_name),
                                "filename": "{}".format(file_name)})


@app.route('/attribute/<attribute>/<hashvalue>')
def get_attribute(attribute, hashvalue):
    with GbdApi(app.config['database']) as gbd_api:
        try:
            values = gbd_api.search(attribute, hashvalue)
            if len(values) == 0:
                return Response("No entry in attribute table associated with this hash", status=404, mimetype="text/plain")
            return str(",".join(str(value) for value in values))
        except ValueError as err:
            return Response("Value Error: {}".format(err), status=500, mimetype="text/plain")


@app.route('/file/<hashvalue>', defaults={'filename': None})
@app.route('/file/<hashvalue>/<filename>')
def get_file(hashvalue, filename):
    with GbdApi(app.config['database']) as gbd_api:
        values = gbd_api.search("local", hashvalue)
        if len(values) == 0:
            return Response("No according file found in our database", status=404, mimetype="text/plain")
        try:
            path = values.pop()
            return send_file(path, as_attachment=True, attachment_filename=os.path.basename(path))
        except FileNotFoundError:
            return Response("Files temporarily not accessible", status=404, mimetype="text/plain")


@app.route('/info/<hashvalue>')
def get_all_attributes(hashvalue):
    with GbdApi(app.config['database']) as gbd_api:
        features = gbd_api.get_features()
        info = dict([])
        for feature in features:
            values = gbd_api.search(feature, hashvalue)
            info.update({feature: str(",".join(str(value) for value in values))})
        return Response(json.dumps(info), status=200, mimetype="application/json")


@app.route("/getdatabase", methods=['GET'])
def get_default_database_file():
    with GbdApi(app.config['database']) as gbd_api:
        app.logger.info('Sending database to {}'.format(request.remote_addr))
        return send_file(gbd_api.get_databases()[0], 
                    as_attachment=True, 
                    attachment_filename=os.path.basename(gbd_api.get_databases()[0]),
                    mimetype='application/x-sqlite3')


def main():
    parser = argparse.ArgumentParser(description='Web- and Micro- Services to access global benchmark database.')
    parser.add_argument('-d', "--db", help='Specify database to work with', default=os.environ.get('GBD_DB'), nargs='?')
    parser.add_argument('-p', "--port", help='Specify port on which to listen', type=int)
    args = parser.parse_args()
    if not args.db:
        eprint("""No database path is given. 
A database path can be given in two ways:
-- by setting the environment variable GBD_DB
-- by giving a path via --db=[path]
A database file containing some attributes of instances used in the SAT Competitions can be obtained at http://gbd.iti.kit.edu/getdatabase
Don't forget to initialize each database with the paths to your benchmarks by using the init-command. """)
    else:
        logging_dir = "gbd-server-logs"
        logging_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), logging_dir)
        if not os.path.exists(logging_path):
            os.makedirs(logging_path)
        logging.basicConfig(filename='{}/server.log'.format(logging_path), level=logging.DEBUG)
        logging.getLogger().addHandler(default_handler)
        global app
        app.wsgi_app = ProxyFix(app.wsgi_app, x_for=1)
        app.config['database'] = args.db
        app.static_folder = os.path.join(os.path.dirname(os.path.abspath(gbd_server.__file__)), "static")
        app.template_folder = os.path.join(os.path.dirname(os.path.abspath(gbd_server.__file__)), "templates-vue")
        app.run(host='0.0.0.0', port=args.port)


if __name__ == '__main__':
    main()



# Global Benchmark Database (GBD)
# Copyright (C) 2020 Markus Iser, Karlsruhe Institute of Technology (KIT)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import multiprocessing
import os
import csv

from multiprocessing import Pool, Lock
from os.path import isfile

from gbd_tool import groups
from gbd_tool import search
from gbd_tool.db import Database
from gbd_tool.gbd_hash import gbd_hash
from gbd_tool.util import eprint, confirm

mutex = Lock()

def import_csv(database, filename, key, source, target, delim_=' '):
    if target not in database.tables():
        print("Target group {} does not exist. Import canceled.".format(target))
    with open(filename, newline='') as csvfile:
        csvreader = csv.DictReader(csvfile, delimiter=delim_, quotechar='\'')
        lst = [(row[key].strip(), row[source].strip()) for row in csvreader if row[source] and row[source].strip()]
        print("Inserting {} values into group {}".format(len(lst), target))
        database.bulk_insert(target, lst)

def add_tag(database, name, value, hash, force=False):
    if database.table_unique(name):
        if force:
            database.submit('REPLACE INTO {} (hash, value) VALUES ("{}", "{}")'.format(name, hash, value))
        else:
            res = database.query("SELECT value FROM {} WHERE hash='{}'".format(name, hash))
            if (len(res) == 0):
                database.submit('INSERT INTO {} (hash, value) VALUES ("{}", "{}")'.format(name, hash, value))
            else:
                existing_value = res[0][0]
                default_value = database.table_default_value(name)
                if existing_value == default_value:
                    eprint("Overwriting default-value {} with new value {} for hash {}".format(default_value, value, hash))
                    database.submit('REPLACE INTO {} (hash, value) VALUES ("{}", "{}")'.format(name, hash, value))
                elif existing_value != value:
                    eprint("Unable to insert tag ({}, {}) into unique '{} (default: {})' as a different value is already set: '{}'".format(hash, value, name, default_value, existing_value))
    else:
        res = database.value_query("SELECT hash FROM {} WHERE hash='{}' AND value='{}'".format(name, hash, value))
        if (len(res) == 0):
            database.submit('INSERT INTO {} (hash, value) VALUES ("{}", "{}")'.format(name, hash, value))


def remove_tags(database, name, hash_list):
    database.submit("DELETE FROM {} WHERE hash IN ('{}')".format(name, "', '".join(hash_list)))

def remove_benchmarks(database):
    eprint("Sanitizing local path entries ... ")
    paths = database.value_query("SELECT value FROM local")
    sanitize = list(filter(lambda path: not isfile(path), paths))
    if len(sanitize) and confirm("{} files not found, remove local path entries from database?".format(len(sanitize))):
        for path in sanitize:
            eprint("File '{}' not found, removing path entry.".format(path))
            database.submit("DELETE FROM local WHERE value='{}'".format(path))

def compute_hash(path):
    eprint('Hashing {}'.format(path))
    hashvalue = gbd_hash(path)
    attributes = [ ('INSERT', 'local', path) ]
    return { 'hashvalue': hashvalue, 'attributes': attributes }

def register_benchmarks(api, database, root, jobs=1):
    pool = Pool(min(multiprocessing.cpu_count(), jobs))
    for root, dirnames, filenames in os.walk(root):
        for filename in filenames:
            path = os.path.join(root, filename)
            if path.endswith(".cnf") or path.endswith(".cnf.gz") or path.endswith(".cnf.lzma") or path.endswith(".cnf.xz") or path.endswith(".cnf.bz2"):
                hashes = database.value_query("SELECT hash FROM local WHERE value = '{}'".format(path))
                if len(hashes) != 0:
                    eprint('Problem {} already hashed'.format(path))
                else:
                    handler = pool.apply_async(compute_hash, args=(path,), callback=api.callback_set_attributes_locked)
                    #handler.get()
    pool.close()
    pool.join() 



# Global Benchmark Database (GBD)
# Copyright (C) 2020 Markus Iser, Luca Springer, Karlsruhe Institute of Technology (KIT)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# python packages
import sqlite3
import multiprocessing

from contextlib import ExitStack
from urllib.error import URLError

# internal packages
from gbd_tool import groups, benchmark_administration, search, bootstrap, sanitize
from gbd_tool.db import Database
from gbd_tool.gbd_hash import gbd_hash
from gbd_tool.util import eprint, is_number


class GbdApi:
    # Create a new GbdApi object which operates on the given databases
    def __init__(self, db_string, jobs=1, separator=" ", inner_separator=",", join_type="INNER"):
        self.databases = db_string.split(":")
        self.jobs = jobs
        self.mutex = multiprocessing.Lock()
        self.separator = separator
        self.inner_separator = inner_separator
        self.join_type = join_type

    def __enter__(self):
        self.database = Database(self.databases)
        with ExitStack() as stack:
            stack.enter_context(self.database)
            self._stack = stack.pop_all()
        return self

    def __exit__(self, exc_type, exc, traceback):
        self._stack.__exit__(exc_type, exc, traceback)

    # Calculate GBD hash
    @staticmethod
    def hash_file(path):
        return gbd_hash(path)

    # Import data from CSV file
    def import_file(self, path, key, source, target):
        benchmark_administration.import_csv(self.database, path, key, source, target, self.separator)

    # Initialize table 'local' with instances found under given path
    def init_database(self, path=None):
        eprint('Initializing local path entries {} using {} cores'.format(path, self.jobs))
        if self.jobs == 1 and multiprocessing.cpu_count() > 1:
            eprint("Activate parallel initialization using --jobs={}".format(multiprocessing.cpu_count()))
        benchmark_administration.remove_benchmarks(self.database)
        benchmark_administration.register_benchmarks(self, self.database, path, self.jobs)

    def bootstrap(self, named_algo):
        bootstrap.bootstrap(self, self.database, named_algo, self.jobs)

    def sanitize(self, hashes):
        sanitize.sanitize(self, self.database, hashes, self.jobs)

    def get_databases(self):
        return self.databases

    # Get all features (or those of given db)
    def get_features(self, path=None):
        if path == None:
            return self.database.tables_and_views()
        elif path in self.databases:
            with Database([path]) as db:
                return db.tables_and_views()
        else:
            return []

    # Get all material features (or those of given db)
    def get_material_features(self, path=None):
        if path == None:
            return self.database.tables()
        elif path in self.databases:
            with Database([path]) as db:
                return db.tables()
        else:
            return []

    # Get all virtual features (or those of given db)
    def get_virtual_features(self, path=None):
        if path == None:
            return self.database.views()
        elif path in self.databases:
            with Database([path]) as db:
                return db.views()
        else:
            return []

    # Check for existence of given feature
    def feature_exists(self, name):
        return name in self.database.tables()

    # Creates the given feature
    def create_feature(self, name, default_value):
        groups.add(self.database, name, default_value is not None, default_value)

    # Removes the given feature
    def remove_feature(self, name):
        groups.remove(self.database, name)

    # Retrieve information about a specific feature
    def get_feature_info(self, name):
        if not name in self.get_features():
            raise ValueError("Attribute '{}' is not available".format(name))
        values = self.database.table_values(name)
        return {'name': name, 
                'unique': self.database.table_unique(name),
                'default-value': self.database.table_default_value(name),
                'num-entries': self.database.table_size(name),
                'numeric-min': values['numeric'][0], 
                'numeric-max': values['numeric'][1], 
                'non-numeric': " ".join(values['discrete']) }

    # Retrieve all values the given feature contains
    def get_feature_values(self, name):        
        if not name in self.get_features():
            raise ValueError("Attribute '{}' is not available".format(name))
        return self.query_search(None, [name], False)

    def callback_set_attributes_locked(self, arg):
        self.set_attributes_locked(arg['hashvalue'], arg['attributes'])

    def set_attributes_locked(self, hash, attributes):
        self.mutex.acquire()
        try:
            # create new connection due to limitations of multi-threaded use (cursor initialization issue)
            with Database(self.databases) as db:
                for attr in attributes:
                    cmd, name, value = attr[0], attr[1], attr[2]
                    db.submit('{} INTO {} (hash, value) VALUES ("{}", "{}")'.format(cmd, name, hash, value))
        finally:
            self.mutex.release()

    # Set the attribute value for the given hashes
    def set_attribute(self, feature, value, hash_list, force):
        if not feature in self.get_material_features():
            raise ValueError("Attribute '{}' is not available (or virtual)".format(feature))
        for h in hash_list:
            benchmark_administration.add_tag(self.database, feature, value, h, force)

    # Remove the attribute value for the given hashes
    def remove_attributes(self, feature, hash_list):
        if not feature in self.get_material_features():
            raise ValueError("Attribute '{}' is not available (or virtual)".format(feature))
        benchmark_administration.remove_tags(self.database, feature, hash_list)

    def search(self, feature, hashvalue):
        if not feature in self.get_features():
            raise ValueError("Attribute '{}' is not available".format(feature))
        return self.database.value_query("SELECT value FROM {} WHERE hash = '{}'".format(feature, hashvalue))

    def hash_search(self, hashes=[], resolve=[], collapse=False, group_by=None):
        try:
            return search.find_hashes(self.database, None, resolve, collapse, group_by, hashes, self.inner_separator, self.join_type)
        except sqlite3.OperationalError as err:
            raise ValueError("Query error for database '{}': {}".format(self.databases, err))

    def query_search(self, query=None, resolve=[], collapse=False, group_by=None):
        try:
            return search.find_hashes(self.database, query, resolve, collapse, group_by, [], self.inner_separator, self.join_type)
        except sqlite3.OperationalError as err:
            raise ValueError("Query error for database '{}': {}".format(self.databases, err))

    def meta_set(self, feature, meta_feature, value):
        self.database.meta_set(feature, meta_feature, value)

    def meta_get(self, feature):
        return self.database.meta_get(feature)

    # clears sepcified meta-features of feature, 
    # or clears all meta-features if meta_feature is not specified
    def meta_clear(self, feature, meta_feature=None):
        self.database.meta_clear(feature, meta_feature)

    def calculate_par2_score(self, query, feature):
        info = self.meta_get(feature)
        if not "timeout" in info:
            eprint("Time-limit 'timeout' missing in meta-record of table '{}'.".format(feature))
            eprint("Unable to calculate score.")
            return
        if not "memout" in info:
            eprint("Memory-limit 'memout' missing in meta-record of table '{}'.".format(feature))
        if not "machine" in info:
            eprint("Machine-id 'machine' missing in meta-record of table '{}'.".format(feature))
        timeout = int(info["timeout"])
        times = self.query_search(query, [feature])
        score = 0
        penalized = set()
        for time in times:
            if is_number(time[1]):
                score += int(time[1])
            else:
                score += 2 * timeout
                penalized.add(time[1])
        print(score/len(times))
        print(penalized)





# Global Benchmark Database (GBD)
# Copyright (C) 2020 Markus Iser, Karlsruhe Institute of Technology (KIT)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sqlite3
import os
import json

from gbd_tool.gbd_hash import HASH_VERSION
from gbd_tool.util import eprint, is_number

VERSION = 0


class DatabaseException(Exception):
    pass


class Database:

    def __init__(self, path_list):
        self.paths = path_list
        # init non-existent databases and check existing databases
        for path in self.paths:
            if not os.path.isfile(path):
                eprint("Initializing DB '{}' with version {} and hash-version {}".format(path, VERSION, HASH_VERSION))
                self.init(path, VERSION, HASH_VERSION)
            else:
                self.check(path, VERSION, HASH_VERSION)

    def __enter__(self):
        #eprint("Main connection: {}".format(self.paths[0]))
        self.connection = sqlite3.connect(self.paths[0])
        self.cursor = self.connection.cursor()
        for path in self.paths[1:]:
            name = os.path.splitext(os.path.basename(path))[0]
            #eprint("Attaching '{}' as {}".format(path, name))
            self.cursor.execute("ATTACH DATABASE '{}' AS {}".format(path, name))
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.commit()
        self.connection.close()

    def init(self, path, version, hash_version):
        con = sqlite3.connect(path)
        cur = con.cursor()
        cur.execute("CREATE TABLE __version (entry UNIQUE, version INT, hash_version INT)")
        cur.execute("INSERT INTO __version (entry, version, hash_version) VALUES (0, {}, {})".format(version, hash_version))
        cur.execute("CREATE TABLE __meta (name TEXT UNIQUE, value BLOB)")
        cur.execute("CREATE TABLE local (hash TEXT NOT NULL, value TEXT NOT NULL)")
        cur.execute("CREATE VIEW IF NOT EXISTS filename (hash, value) AS SELECT hash, REPLACE(value, RTRIM(value, REPLACE(value, '/', '')), '') FROM local")
        con.commit()
        con.close()

    def check(self, path, version, hash_version):
        con = sqlite3.connect(path)
        cur = con.cursor()
        lst = cur.execute("SELECT tbl_name FROM sqlite_master WHERE type='table'")
        tables = [x[0] for x in lst]
        if not "__version" in tables:
            eprint("WARNING: Version info not available in database {}".format(path))
            return
        __version = cur.execute("SELECT version, hash_version FROM __version").fetchall()
        if __version[0][0] != version:
            eprint("WARNING: DB Version is {} but tool version is {}".format(__version[0][0], version))
        if __version[0][1] != hash_version:
            eprint("WARNING: DB Hash-Version is {} but tool hash-version is {}".format(__version[0][1], hash_version))

        # upgrade legacy data-model
        if "filename" in tables:
            cur.execute("DROP TABLE IF EXISTS filename")        
            cur.execute("CREATE VIEW IF NOT EXISTS filename (hash, value) AS SELECT hash, REPLACE(value, RTRIM(value, REPLACE(value, '/', '')), '') FROM local")
            con.commit()

        if not "__meta" in tables:
            cur.execute("CREATE TABLE __meta (name TEXT UNIQUE, value BLOB)")

        con.close()

    def value_query(self, q):
        lst = self.cursor.execute(q).fetchall()
        return set([row[0] for row in lst])

    def query(self, q):
        return self.cursor.execute(q).fetchall()

    def submit(self, q):
        eprint(q)
        self.cursor.execute(q)
        self.commit()

    def bulk_insert(self, table, lst):
        if self.table_unique(table):
            self.cursor.executemany("REPLACE INTO {} VALUES (?,?)".format(table), lst)
        else:
            self.cursor.executemany("INSERT INTO {} VALUES (?,?)".format(table), lst)

    def commit(self):
        self.connection.commit()

    def tables_and_views(self):
        lst = self.query(r"SELECT tbl_name FROM sqlite_master WHERE (type='table' OR type='view') AND NOT tbl_name LIKE '\_\_%' escape '\' AND NOT tbl_name LIKE 'sqlite\_%' escape '\'")
        return [x[0] for x in lst]

    def tables(self):
        lst = self.query(r"SELECT tbl_name FROM sqlite_master WHERE type='table' AND NOT tbl_name LIKE '\_\_%' escape '\' AND NOT tbl_name LIKE 'sqlite\_%' escape '\'")
        return [x[0] for x in lst]

    def views(self):
        lst = self.query(r"SELECT tbl_name FROM sqlite_master WHERE type='view' AND NOT tbl_name LIKE '\_\_%' escape '\' AND NOT tbl_name LIKE 'sqlite\_%' escape '\'")
        return [x[0] for x in lst]

    def table_info(self, table):
        lst = self.query("PRAGMA table_info({})".format(table))
        columns = ('index', 'name', 'type', 'notnull', 'default_value', 'pk')
        table_infos = [dict(zip(columns, values)) for values in lst]
        return table_infos

    def index_list(self, table):
        lst = self.query("PRAGMA index_list({})".format(table))
        columns = ('seq', 'name', 'unique', 'origin', 'partial')
        index_list = [dict(zip(columns, values)) for values in lst]
        return index_list
    
    def index_info(self, index):
        tup = self.query("PRAGMA index_info({})".format(index))
        columns = ('index_rank', 'table_rank', 'name')
        index_info = dict(zip(columns, tup[0]))
        return index_info

    def table_info_augmented(self, table):
        table_infos = [info.update({'unique': False}) or info for info in self.table_info(table)]
        
        # determine unique columns
        index_list = self.index_list(table)
        for index in [e for e in index_list if e['unique']]:
            col = self.index_info(index['name'])['table_rank']
            table_infos[col]['unique'] = True

        for info in table_infos:
            if info['default_value'] is not None:
                info['default_value'] = info['default_value'].strip('"')
        
        return table_infos

    # return list of distinct values and value-range for numeric values
    def table_values(self, table):
        result = { "numeric" : [None, None], "discrete" : [] }
        values = self.value_query('SELECT DISTINCT value FROM {}'.format(table))
        for value in values:
            if is_number(value):
                if result["numeric"][0] == None:
                    result["numeric"] = [value, value]
                elif value < result["numeric"][0]:
                    result["numeric"][0] = value
                elif value > result["numeric"][1]:
                    result["numeric"][1] = value
            else:
                result["discrete"].append(value)
        return result

    def table_size(self, table):
        return self.value_query('SELECT COUNT(*) FROM {}'.format(table)).pop()

    def table_unique(self, table):
        return self.table_info_augmented(table)[0]['unique']

    def table_default_value(self, table):
        return self.table_info_augmented(table)[1]['default_value']

    def meta_get(self, table):
        return json.loads((self.value_query("SELECT value FROM __meta WHERE name = '{}'".format(table)) or {"{}"}).pop())

    def meta_clear(self, table, meta_feature=None):
        if not meta_feature:
            self.submit("INSERT OR REPLACE INTO __meta (name, value) VALUES ('{}', '')".format(table))
        else:
            values = self.meta_get(table)
            if meta_feature in values:
                values.pop(meta_feature)
            self.submit("INSERT OR REPLACE INTO __meta (name, value) VALUES ('{}', '{}')".format(table, json.dumps(values)))

    def meta_set(self, table, meta_feature, value):
        values = self.meta_get(table)
        values[meta_feature] = value
        self.submit("INSERT OR REPLACE INTO __meta (name, value) VALUES ('{}', '{}')".format(table, json.dumps(values)))





# Global Benchmark Database (GBD)
# Copyright (C) 2020 Markus Iser, Karlsruhe Institute of Technology (KIT)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gbd_tool.util import eprint
from tatsu import parse, exceptions
import pprint


def find_hashes(database, query=None, resolve=[], collapse=False, group_by=None, hashes=[], separator=",", join_type="INNER"):
    statement = "SELECT {} FROM local {} WHERE {} GROUP BY {}"
    s_attributes = "local.hash"
    s_group_by = 'local.hash'
    s_tables = ""
    s_conditions = "1=1"
    tables = { "local" }
    if resolve is None:
        resolve = []
    
    if query is not None and query:
        try:
            ast = parse(GRAMMAR, query)
        except exceptions.FailedParse as err:
            eprint(err)
            eprint("Exception in Query-Parser: Put any arithmetic expression in parentheses.")
            return list() 
        s_conditions = build_where(ast)
        tables.update(collect_tables(ast))
    elif len(hashes) > 0:
        s_conditions = "local.hash in ('{}')".format("', '".join(hashes))

    if group_by is not None:
        s_group_by = group_by + ".value"
        resolve.append(group_by)

    if collapse:
        s_attributes = "MIN(DISTINCT(local.hash))"
        if len(resolve):
            s_attributes = s_attributes + ", " + ", ".join(['MIN(DISTINCT({}.value))'.format(table) for table in resolve])
    else:
        s_attributes = 'REPLACE( GROUP_CONCAT(DISTINCT(local.hash)), ",", "{}" )'.format(separator)
        if len(resolve):
            s_attributes = s_attributes + ", " + ", ".join(['REPLACE( GROUP_CONCAT(DISTINCT({}.value)), ",", "{}" )'.format(table, separator) for table in resolve])
    tables.update(resolve)

    s_tables = " ".join(['{} JOIN {} ON local.hash = {}.hash'.format(join_type, table, table) for table in tables if table != "local"])

    eprint(statement.format(s_attributes, s_tables, s_conditions, s_group_by))

    return database.query(statement.format(s_attributes, s_tables, s_conditions, s_group_by))


def build_where(ast):
    result = ""
    if ast["q"] is not None:
        result = build_where(ast["q"])
    elif ast["qop"] is not None:
        result = '(' + build_where(ast["left"]) + " " + ast["qop"] + " " + build_where(ast["right"]) + ')'
    elif ast["sop"] is not None:
        result = ast["left"] + ".value " + ast["sop"] + " \"" + ast["right"] + "\""
    elif ast["aop"] is not None:
        result = build_where(ast["left"]) + " " + ast["aop"] + " " + build_where(ast["right"])
    elif ast["bracket_term"] is not None:
        result = '(' + build_where(ast["bracket"]) + ')'
    elif ast["top"] is not None:
        result = build_where(ast["left"]) + " " + ast["top"] + " " + build_where(ast["right"])
    elif ast["value"] is not None:
        result = "CAST(" + ast["value"] + ".value AS FLOAT)"
    elif ast["constant"] is not None:
        result = ast["constant"]
    return result


def collect_tables(ast):
    result = set()
    if ast["q"] is not None:
        result.update(collect_tables(ast["q"]))
    elif ast["qop"] is not None:
        result.update(collect_tables(ast['left']))
        result.update(collect_tables(ast['right']))
    elif ast["sop"] is not None:
        result.add(ast["left"])
    elif ast["aop"] is not None:
        result.update(collect_tables(ast["left"]))
        result.update(collect_tables(ast["right"]))
    elif ast["bracket_term"] is not None:
        result.update(collect_tables(ast["bracket_term"]))
    elif ast["top"] is not None:
        result.update(collect_tables(ast['left']))
        result.update(collect_tables(ast['right']))
    elif ast["value"] is not None:
        result.add(ast["value"])
    return result


GRAMMAR = r'''
    @@grammar::EXP
    @@ignorecase::True

    start = q:query $ ;

    query = '(' q:query ')' | left:query qop:('and' | 'or') right:query | scon | acon;

    scon = left:colname sop:('=' | '!=') right:alnum | left:colname sop:('like') right:likean ;
        
    acon = left:term aop:('=' | '!=' | '<' | '>' | '<=' | '>=' ) right:term ;

    term = value:colname | constant:num | '(' left:term top:('+'|'-'|'*'|'/') right:term ')' ;

    num = /[0-9\.\-]+/ ;
    alnum = /[a-zA-Z0-9_\.\-\/\?]+/ ;
    likean = /[\%]?[a-zA-Z0-9_\.\-\/\?]+[\%]?/;
    colname = /[a-zA-Z][a-zA-Z0-9_]+/ ;
'''



# Global Benchmark Database (GBD)
# Copyright (C) 2020 Markus Iser, Karlsruhe Institute of Technology (KIT)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gbd_tool.util import eprint

def add(database, name, unique=False, default=None):
    ustr = "UNIQUE" if unique else ""
    dstr = "DEFAULT \"{}\"".format(default) if default is not None else ""
    database.submit(
        'CREATE TABLE IF NOT EXISTS {} (hash TEXT {} NOT NULL, value TEXT NOT NULL {})'.format(name, ustr, dstr))
    if default is not None:
        database.submit('INSERT OR IGNORE INTO {} (hash) SELECT hash FROM local'.format(name))
        database.submit('''CREATE TRIGGER {}_dval AFTER INSERT ON local BEGIN 
                INSERT INTO {} (hash) VALUES (NEW.hash); END'''.format(name, name))

def remove(database, name):
    database.submit('DROP TABLE IF EXISTS {}'.format(name))
    database.submit('DROP TRIGGER IF EXISTS {}_dval'.format(name))
